(* $Id: cmd.ml 426 2004-05-07 22:00:28Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Printf


let usage() =
  prerr_endline ("usage: godi_console [<command> [<options>]]\n" ^
		 "Without command: start interactive console\n" ^
		 "Available commands:\n" ^
		 "  update - Update the build instructions\n" ^
		 "  wish - Manipulate the list of user's wishes\n" ^
		 "  perform - Install or remove packages\n" ^
		 "  list - List packages by criterion\n" ^
		 "The commands have their own help texts, use -help\n" ^
		 "to see them, e.g.\n" ^
		 "  godi_console perform -help");
  exit 1
;;

let opt_re = Str.regexp "^\\([^=:]+\\):\\([^=]+\\)=\\(.*\\)$";;
let ropt_re = Str.regexp "^\\([^=:]+\\):\\([^=]+\\)$";;

let split_re = Str.regexp "[ \t\r\n]+";;
let split s = Str.split split_re s;;

let word_by_word f s =
  let l = split s in
  List.iter f l
;;


let cmd_update() =
  Arg.current := 1;
  Arg.parse 
      []
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      "usage: godi_console update [<options>]";
  Godi_update.update_available();
  Godi_update.update_build_instructions()
;;


let cmd_wish() =
  let wishes = ref (Godi_wishes.read_wishes ~pref_user_error:true ()) in
  let modified = ref false in

  let check_src_pkg pkgname =
    let src_pkgs = Godi_query.scan_source_packages() in
    if not (List.exists (fun p -> pkg_name p = pkgname) src_pkgs) then
      user_error ("No such source package: " ^ pkgname)
  in

  let check_inst_pkg pkgname =
    let inst_pkgs = Godi_query.scan_installed_packages() in
    if not (List.exists (fun p -> pkg_name p = pkgname) inst_pkgs) then
      user_error ("No such installed package: " ^ pkgname)
  in

  let add_build_wish pkgname =
    check_src_pkg pkgname;
    let wish = `Build pkgname in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
      modified := true
    )
  in
 
  let build_all_installed() =
    let src_pkgs = Godi_query.scan_source_packages() in
    let src_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) src_pkgs) in
    let inst_pkgs = Godi_query.scan_installed_packages() in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let spkgs = PkgMap.find_by ~name:n src_pkgs_map in
	 if spkgs <> [] then add_build_wish n
      )
      inst_pkgs
  in

  let rem_build_wish pkgname =
    (* check_src_pkg pkgname; *)
    let wish = `Build pkgname in
    wishes := List.filter (fun w -> w <> wish) !wishes;
    modified := true
  in
    
  let add_remove_wish pkgname =
    check_inst_pkg pkgname;
    let wish = `Remove pkgname in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
      modified := true
    );
  in
 
  let rem_remove_wish pkgname =
    (* check_inst_pkg pkgname; *)
    let wish = `Remove pkgname in
    wishes := List.filter (fun w -> w <> wish) !wishes;
    modified := true
  in
    
  let add_option opt =
    if Str.string_match opt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let optvalue = Str.matched_group 3 opt in
      check_src_pkg pkgname;
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl @ [ `Option(pkgname,optname,optvalue) ];
      modified := true
    )
    else
      user_error ("Syntax error in -option " ^ opt)
  in
    
  let rem_option opt =
    if Str.string_match ropt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl;
      modified := true
    )
    else
      user_error ("Syntax error in -no-option " ^ opt)
  in

  let reset() =
    wishes := [];
    modified := true
  in

  let print() =
    (* Sort wishes: *)
    let pkg_of_wish =
      function
	  `Build p -> p
	| `Remove p -> p
	| `Option(p,_,_) -> p in
    let cat_of_wish =
      function
	  `Build _ -> 1
	| `Remove _ -> 0
	| `Option(_,_,_) -> 2 in
    let optname_of_wish =
      function
	  `Build _ -> ""
	| `Remove _ -> ""
	| `Option(n,_,_) -> n in
    let cmp w1 w2 =
      let p1 = pkg_of_wish w1 in
      let p2 = pkg_of_wish w2 in
      match compare p1 p2 with
	  0 ->
	    ( match cat_of_wish w1 - cat_of_wish w2 with
		  0 ->
		    compare (optname_of_wish w1) (optname_of_wish w2)
		| n -> n
	    )
	| n -> n
    in
    let wl = List.sort cmp !wishes in
    (* Finally print wishes: *)
    List.iter
      (function
	  `Build p ->
	    printf "-build  %s\n" p
	| `Remove p ->
	    printf "-remove %s\n" p
	| `Option(p,n,v) ->
	    printf "-option %s:%s=%s\n" p n v
      )
      wl
  in

  Arg.current := 1;
  Arg.parse 
      [ "-build", Arg.String (word_by_word add_build_wish),
	       "<pkg>                 Add wish to build <pkg>";
	"-no-build", Arg.String (word_by_word rem_build_wish),
	          "<pkg>              Withdraw wish to build <pkg>";
	"-remove", Arg.String (word_by_word add_remove_wish),
	        "<pkg>                Add wish to remove <pkg>";
	"-no-remove", Arg.String (word_by_word rem_remove_wish),
	           "<pkg>             Withdraw wish to remove <pkg>";
	"-option", Arg.String add_option,
	        "<pkg>:<name>=<val>   Add option <name>=<val> for <pkg>";
	"-no-option", Arg.String rem_option,
	           "<pkg>:<name>      Withdraw option <name> for <pkg>";
	"-rebuild", Arg.Unit build_all_installed,
	         "                    Wish to rebuild all installed packages";
	"-reset", Arg.Unit reset,
	       "                      Withdraw all wishes";
	"-print", Arg.Unit print,
	       "                      Print resulting wish list";
      ]
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      "usage: godi_console wish <options>";

  if !modified then
    Godi_wishes.write_wishes ~pref_user_error:true !wishes
;;


let cmd_perform() =
  let wishes = ref [] in
  let use_wishlist = ref false in
  let dry_run = ref false in
  let build_all_installed = ref false in
  let newer = ref false in

  let check_src_pkg pkgname =
    let src_pkgs = Godi_query.scan_source_packages() in
    if not (List.exists (fun p -> pkg_name p = pkgname) src_pkgs) then
      user_error ("No such source package: " ^ pkgname)
  in

  let check_inst_pkg pkgname =
    let inst_pkgs = Godi_query.scan_installed_packages() in
    if not (List.exists (fun p -> pkg_name p = pkgname) inst_pkgs) then
      user_error ("No such installed package: " ^ pkgname)
  in

  let add_build_wish pkgname =
    check_src_pkg pkgname;
    let wish = `Build pkgname in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
    )
  in
 
  let add_remove_wish pkgname =
    check_inst_pkg pkgname;
    let wish = `Remove pkgname in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
    );
  in
 
  let add_option opt =
    if Str.string_match opt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let optvalue = Str.matched_group 3 opt in
      check_src_pkg pkgname;
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl @ [ `Option(pkgname,optname,optvalue) ];
    )
    else
      user_error ("Syntax error in -option " ^ opt)
  in

  let add_wishes() =
    use_wishlist := true
  in

  Arg.current := 1;
  Arg.parse 
      [ "-build", Arg.String (word_by_word add_build_wish),
	       "<pkg>                 Build <pkg>";
	"-remove", Arg.String (word_by_word add_remove_wish),
	        "<pkg>                Remove <pkg>";
	"-option", Arg.String add_option,
	        "<pkg>:<name>=<val>   Set option <name>=<val> for <pkg>";
	"-wishes", Arg.Unit add_wishes,
	        "                     Build/remove packages of the wish list";
	"-rebuild", Arg.Set build_all_installed,
	         "                    Rebuild all installed packages";
	"-newer", Arg.Set newer,
	       "                      Build only src packages newer than inst counterparts";
	"-dry-run", Arg.Set dry_run,
	         "                    Only generate a plan but don't carry it out";
      ]
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      "usage: godi_console perform <options>";

  (* First scan packages *)
  let src_pkgs = Godi_query.scan_source_packages() in
  let inst_pkgs = Godi_query.scan_installed_packages() in

  (* Apply wishes *)
  Godi_wishes.apply_wishes_to_packages src_pkgs !wishes;
  Godi_wishes.apply_wishes_to_packages inst_pkgs !wishes;

  let persistent_wl = Godi_wishes.read_wishes ~pref_user_error:true () in
  if !use_wishlist then (
    Godi_wishes.apply_wishes_to_packages src_pkgs persistent_wl;
    Godi_wishes.apply_wishes_to_packages inst_pkgs persistent_wl;
  );
 
  if !build_all_installed then (
    let src_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) src_pkgs) in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let spkgs = PkgMap.find_by ~name:n src_pkgs_map in
	 List.iter
	   (fun (spkg,_) ->
	      (source_pkg spkg)#set_user_status `Build
	   )
	   spkgs
      )
      inst_pkgs
  );

  if !newer then (
    let inst_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) inst_pkgs) in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let ipkgs = PkgMap.find_by ~name:n inst_pkgs_map in
	 match ipkgs with
	     [] ->
	       ()
	   | [ipkg,_] ->
	       (* Is pkg newer than ipkg? *)
	       if not(newer_package pkg ipkg) then (
		 (* pkg is not newer, so do not build it *)
		 (source_pkg pkg)#set_user_status `Keep
	       )
	   | _ ->
	       (* Several installed packages with the same name: impossible *)
	       assert false
      )
      src_pkgs
  );

  (* Prepare a function to write wishes back to the file: *)
  let save_wishes() =
    if !use_wishlist then (
      let wl = Godi_wishes.wishes_of_packages(src_pkgs @ inst_pkgs) in
      (* We only write those wishes that are also contained in 
       * persistent_wl
       *)
      let wl' =
	List.filter (fun w -> List.mem w persistent_wl) wl in
      Godi_wishes.write_wishes ~pref_user_error:true wl';
    )
  in

  (* Write a temporary godi.conf file: *)
  let tmp_godi_conf =  Filename.temp_file "tmpgodi" ".conf" in
  let mflag = Godi_conffile.modify_godi_conf
                ~pref_user_error:true ~destfile:tmp_godi_conf src_pkgs in
 
  (* When options have been modified, rescan source packages: *)
  let vid = new_var_id() in
  let src_pkgs' =
    if mflag then (
      List.iter (fun pkg -> 
		   (source_pkg pkg)#rescan_deps tmp_godi_conf vid)
		src_pkgs;
      List.map 
	(fun pkg -> `Source_pkg ((source_pkg pkg)#variant vid))
	src_pkgs
    )
    else src_pkgs in
  Sys.remove tmp_godi_conf;

  (* Generate the plan: *)
  let n = Godi_plan.update_plan ~iteration:1 (src_pkgs' @ inst_pkgs) in
  (* May raise Mission_impossible *)
 
  let to_remove =
    List.filter to_be_removed inst_pkgs in

  let to_remove' =
    List.rev(Godi_plan.sort_packages to_remove) in
 
  let to_build =
    List.filter to_be_built src_pkgs' in

  let to_build' =
    Godi_plan.sort_packages to_build in

 
  (* Execute the plan: *)

  if to_remove = [] && to_build = [] && not mflag then
    print_endline "Nothing to do."
  else (
    List.iter
	  (fun pkg ->
	     printf "### Fetching %s (if necessary)\n" (pkg_name pkg);
	     flush stdout;
	     if not !dry_run then
	       Godi_perform.fetch_pkg ~pref_user_error:true pkg;
	   )
	   to_build';
    List.iter
	  (fun pkg ->
	     if (installed_pkg pkg)#preserve then
	       printf "### Package %s cannot be removed, skipping\n" (pkg_name pkg)
	     else (
	       printf "### Removing %s\n" (pkg_name pkg);
	       flush stdout;
	       if not !dry_run then (
		 Godi_perform.remove_pkg ~pref_user_error:true pkg;
		 (installed_pkg pkg)#set_user_status `Keep;
		 save_wishes();
	       )
	     )
	  )
	  to_remove';

    if mflag then (
      printf "### Modifying godi.conf\n";
      flush stdout;
      if not !dry_run then (
	let _ = Godi_conffile.modify_godi_conf ~pref_user_error:true to_build' in
	()
      )
    );

    List.iter
      (fun pkg ->
	 printf "### Building %s\n" (pkg_name pkg);
	 flush stdout;
	 if not !dry_run then (
	   try
	     Godi_perform.build_pkg ~pref_user_error:true pkg;
	     (source_pkg pkg)#set_user_status `Keep;
	     save_wishes();
	   with
	       Restart_recommended sh_cmd ->
		 printf "### The installation of the package requires a restart of this program!\n";
		 flush stdout;
		 (* Remove the wish, if any. Note that we do not know whether
		  * sh_cmd will be successful, so this may be too early.
		  *)
		 (source_pkg pkg)#set_user_status `Keep;
		 save_wishes();
		 (* Determine the remaining packages to build: *)
		 let to_do = ref [] in
		 let found = ref false in
		 List.iter (fun pkg' -> 
			      if !found then to_do := pkg' :: !to_do;
			      if pkg = pkg' then found := true) to_build';
		 let restart_args =
		   List.map
		     (fun pkg' ->
			"-build " ^ Filename.quote (pkg_name pkg'))
		     (List.rev !to_do) in
		 let restart_cmd =
		   "exec " ^ Filename.quote Sys.argv.(0) ^ " perform " ^ 
		   String.concat " " restart_args in
		 raise(Restart(sh_cmd ^ "\n" ^ restart_cmd));
		   
	 )
      )
      to_build'
  )
;;


let print_deps l =
  String.concat ", "
    (List.map Godi_versions.print_dependency l)
;;


let indent s =
  Str.global_replace (Str.regexp "\n") "\n    " s
;;


let remove_linefeeds s =
  Str.global_replace (Str.regexp "[\r\n]") " " s
;;


let version_str pkg =
  let v = pkg_version pkg in
  let r = pkg_pkgrevision pkg in
  if r > 0 then v ^ "godi" ^ string_of_int r else v
;;


let cmd_list() =
  let list_src = ref false in
  let list_inst = ref false in
  let only_names = ref false in
  let long = ref false in
  let filters = ref [] in

  Arg.current := 1;
  Arg.parse 
      [ "-source", Arg.Set list_src,
                "           List packages available as source code";
	"-installed", Arg.Set list_inst,
	           "        List installed packages";
	"-names", Arg.Set only_names,
	       "            Output the pure list of package names (no versions etc.)";
	"-long", Arg.Set long,
	      "             Output descriptions, options, etc.";
	
      ]
      (fun s -> 
	 let dep = Godi_versions.parse_dependency s in
	 filters := !filters @ [dep])
      "usage: godi_console list [<options>] [<pkgfilter> ...]";

  if not !list_src && not !list_inst then (
    list_src := true;
    list_inst := true;
  );

  (* First scan packages *)
  let src_pkgs = Godi_query.scan_source_packages() in
  let inst_pkgs = Godi_query.scan_installed_packages() in
  let out_pkgs = src_pkgs @ inst_pkgs in

  (* filters: Select which packages are printed *)
  let is_selected pkg =
    (!filters = []) ||
    ( (* Package is selected when there is a matching filter *)
      List.exists
	(fun (n,cond) ->
	   (n = (pkg_name pkg)) &&
	   (Godi_versions.version_matches_condition
	      (pkg_version pkg) cond))
	!filters
    )
  in

  (* Arrange packages by name: *)
  let h = Hashtbl.create 20 in
  let keys = ref [] in
  List.iter
    (fun pkg ->
       let n = pkg_name pkg in
       let old =
	 try Hashtbl.find h n with Not_found -> [] in
       if old = [] then keys := n :: !keys;
       Hashtbl.replace h n (pkg :: old)
    )
    out_pkgs;

  keys := List.sort compare !keys;

  (* Output list *)
  List.iter
    (fun name ->
       let pkgs = Hashtbl.find h name in
       let (inst_version, inst_sel) =
	 try
	   let inst_pkg = 
	     List.find (fun pkg -> pkgtype pkg = `Installed) pkgs in
	   ( Some(version_str inst_pkg),
	     !list_inst && is_selected inst_pkg
	   )
	 with
	     Not_found -> (None, false) in
       let (src_version, src_sel) =
	 try
	   let src_pkg = 
	     List.find (fun pkg -> pkgtype pkg = `Source) pkgs in
	   ( Some(version_str src_pkg),
	     !list_src && is_selected src_pkg
	   )
	 with
	     Not_found -> (None, false) in

       let sel = inst_sel || src_sel in
       
       if sel then (
	 if !only_names then
	   print_endline name
	 else (
	   let inst_version_str =
	     match inst_version with
		 None -> "-"
	       | Some v -> v in
	   let src_version_str =
	     match src_version with
		 None -> "-"
	       | Some v -> v in

	   (* Long format:
	    * - The package name is on column 0
	    * - The attributes begin on column 2
	    * - Continuation lines begin on column 4
	    * ==> Easy to parse!
	    *)

	   if !long then (
	     List.iter
               (fun pkg ->			    
		  let sel =
		    match pkgtype pkg with
			`Source -> !list_src
		      | `Installed -> !list_inst
		      | _ -> false in
		  if sel then (
		    printf "%s\n" name;
		    ( match pkgtype pkg with
			  `Source -> 
			    printf "  type: source\n"
			| `Installed ->
			    printf "  type: installed\n"
			| _ -> assert false
		    );
		    printf "  version: %s\n" (version_str pkg);
		    printf "  comment: %s\n" (remove_linefeeds (pkg_comment pkg));
		    printf "  description: %s\n" (indent (pkg_description pkg));
		    printf "  dependencies: %s\n" (print_deps
						     (pkg_dependencies pkg));
		    printf "  build_deps: %s\n" (print_deps
						   (pkg_build_dependencies pkg));
		    ( match pkgtype pkg with
			  `Source ->
			    printf "  build_message: %s\n"
			      (indent (source_pkg pkg)#build_message);
			    let nv = (source_pkg pkg)#values_of_build_options in
			    printf "  build_options: %s\n"
			      (indent
				 (String.concat "\n" 
				    (List.map
				       (fun (n,v) -> n ^ "=" ^ v)
				       nv)));
			| `Installed ->
			    printf "  preserved: %b\n"
			      (installed_pkg pkg)#preserve;
			| _ ->
			    ()
		    );
		    printf "\n";
		  )
	       )
	       pkgs

	   )
	   else (
	     printf "%-30s inst:%-10s src:%-10s\n"
		   name
		   inst_version_str
		   src_version_str
	   )
	 )
       )
    )
    !keys;
;;


let exec_cmd() =
  try
    ( let n = Array.length Sys.argv in
      if n < 2 then usage();
      match Sys.argv.(1) with
	  "update"  -> cmd_update()
	| "wish"    -> cmd_wish()
	| "perform" -> cmd_perform()
	| "list"    -> cmd_list()
	| _ ->
	    usage()
    )
  with
    | Godi_plan.Mission_impossible text ->
	prerr_endline "There is no solution because of conflicting dependencies.";
	prerr_endline text;
	exit 1
    | User_error msg ->
	prerr_endline ("Error: " ^ msg);
	exit 1
;;
