(* $Id: menu_install.ml 426 2004-05-07 22:00:28Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Printf

let help_str =
  "This dialog appears if the build/removal/upgrade plan is consistent, and\n" ^
  "can be actually executed. Press 'o'=ok to start execution.\n" ^
  "\n" ^
  "The next screen will show lots of messages output by the build tools\n" ^
  "GODI invokes. This is a protocol of all performed actions, and it may\n" ^
  "be very helpful for analysis if something goes wrong. (You can save it\n" ^
  "into a file with the 's'=save key.)\n" ^
  "\n" ^
  "You can abort the build process by pressing 'a'.\n" ^ 
  "\n" ^
  "When the build process is over, a line like\n" ^
  "\n" ^
  "  ### Installation completed\n" ^
  "\n" ^
  "or\n" ^
  "\n" ^
  "  ### Command failed\n" ^
  "\n" ^
  "is appended to the protocol to inform you about the success of the\n" ^
  "process."
;;


let help_active_str =
  "The build process is currently active, and the messages printed on the\n" ^
  "screen document the progress of the process.\n" ^
  "\n" ^
  "You can abort the build process by pressing 'a'.\n" ^ 
  "\n" ^
  "When the build process is over, a line like\n" ^
  "\n" ^
  "  ### Installation completed\n" ^
  "\n" ^
  "or\n" ^
  "\n" ^
  "  ### Command failed\n" ^
  "\n" ^
  "is appended to the protocol to inform you about the success of the\n" ^
  "process."
;;


let help_over_str =
  "The build process is over. Check the last line for a success message." 
;;


let help_self_update_str =
  "The current package is a very special one: It contains the replacement\n" ^
  "for the running GODI console. Because some operating systems cannot\n" ^
  "replace programs that are in running state, special actions must be\n" ^
  "performed to handle this situation. These actions have been automated,\n" ^
  "so no special user interaction is required, but you will notice that\n" ^
  "the console stops, that shell commands will be executed, and that finally\n" ^
  "the console will be restarted. Of course, the build process will be resumed,\n" ^
  "as there might be packages that need still to be installed.\n" ^ 
  "\n" ^ 
  "Summarized, some strange things will now happen on your terminal, but\n" ^
  "this is alright.\n" ^
  "\n" ^
  "Technical information: the .tgz package file have already been created,\n" ^
  "and needs to be installed. This is done by executing godi_delete and\n" ^
  "godi_add statements from the shell. These shell commands have been prepared\n" ^
  "and are now executed as substitute for the console. Finally, the console\n" ^
  "executable is started again; of course, this is the new version of the console\n" ^
  "that has just been installed (another reason to restart the console)."
;;


let count_remove corr_list =
  let list_rm = 
    List.filter
      (fun (n, srcl, binl, instl) ->
	 List.exists to_be_removed instl &&
	 not(List.exists to_be_built srcl) 
	 (* TODO: check binl, too *)
      )
      corr_list in
  List.length list_rm
;;


let count_new_build corr_list =
  let list_bld = 
    List.filter
      (fun (n, srcl, binl, instl) ->
	 List.exists to_be_built srcl &&
	 not(List.exists to_be_removed instl)
	 (* TODO: check binl, too *)
      )
      corr_list in
  List.length list_bld
;;


let count_upgrade corr_list =
  let list_upg = 
    List.filter
      (fun (n, srcl, binl, instl) ->
	 List.exists to_be_built srcl &&
	 List.exists to_be_removed instl
	 (* TODO: check binl, too *)
      )
      corr_list in
  List.length list_upg
;;


let lf_re = Str.regexp "\n";;


let forget_src_pkgs_on_quit = ref false;;


let execute (pg : Miniui.page) src_list inst_list =
  
  let poll () =
    pg # update();
    pg # poll_event()
  in

  let write_wishes() =
    let wl = Godi_wishes.wishes_of_packages(src_list @ inst_list) in
    Godi_wishes.write_wishes ~pref_user_error:true wl;
  in

  pg # clear();
  pg # configure ~auto_scroll:true ();

  Menu_aux.add_compact_head ~lmargin:6 pg ">>> Installation <<<";

  let active = ref true in
               (* set this to false to stop the installation process *)
  let disp_proto = new Protocol.display_protocol pg in
  let protocol = disp_proto # collector in
  let ia = { Godi_exec.period = 0.2;
	     Godi_exec.callback = (fun () ->
				     poll();
				     !active);
	   } in

  let cancel_btn =
    pg # add_foot_button 
      ~text:"abort"
      ~underline:0
      ~callback:(fun () ->
		   active := false)
      (`Keypress 'a') in

  let help_btn =
    pg # add_foot_button ~text:"help" ~underline:0
      ~callback:(Menu_help.display pg#new_page help_active_str)
      (`Keypress 'h') in

  let ui_continue() =
    pg # remove_button cancel_btn;
    pg # remove_button help_btn;
    ignore(pg # add_foot_button
	     ~text:"save protocol"
	     ~underline:0
	     ~callback:(fun () ->
			  disp_proto # save_dialog())
	     (`Keypress 's'));
    ignore(pg # add_foot_button ~text:"help" ~underline:0
	     ~callback:(Menu_help.display pg#new_page help_over_str)
	     (`Keypress 'h'));

    ignore(pg # add_foot_button 
	     ~text:"exit menu"
	     ~underline:1
	     ~callback:(fun () -> pg#quit())
	     (`Keypress 'x'));


    pg # loop();
    pg # quit();  (* also terminate the loop() called from [display] below *)
  in

  let ui_self_update sh_cmd =
    pg # remove_button cancel_btn;
    pg # remove_button help_btn;
    ignore(pg # add_foot_button
	     ~text:"save protocol"
	     ~underline:0
	     ~callback:(fun () ->
			  disp_proto # save_dialog())
	     (`Keypress 's'));
    ignore(pg # add_foot_button ~text:"help" ~underline:0
	     ~callback:(Menu_help.display pg#new_page help_self_update_str)
	     (`Keypress 'h'));

    ignore(pg # add_foot_button 
	     ~text:"ok"
	     ~underline:0
	     ~callback:(fun () -> 
			  (* Raise an exception that will execute sh_cmd,
			   * and restart the installation procedure:
			   *)
			  let self_binary = Filename.quote Sys.argv.(0) in
			  raise(Restart(sh_cmd ^ "\n" ^
					"exec " ^ self_binary ^ " interactive -continue-installation")))
	     (`Keypress 'o'));

    ignore(pg # add_foot_button 
	     ~text:"abort"
	     ~underline:0
	     ~callback:(fun () -> pg#quit())
	     (`Keypress 'a'));


    pg # loop();
    pg # quit();  (* also terminate the loop() called from [display] below *)
  in

  (* Extract the packages to remove and to build: *)

  let to_remove =
    List.filter to_be_removed inst_list in

  let to_build =
    List.filter to_be_built src_list in

  (* Change all `Build_dep packages into `Build packages (because the
   * user has confirmed that!). Otherwise, dependencies would be lost
   * if the build was interrupted
   *)

  List.iter
    (fun pkg ->
       (any_pkg pkg) # set_user_status `Build
    )
    to_build;

  try

    write_wishes();

    let to_remove' =
      List.rev(Godi_plan.sort_packages to_remove) in
    let to_build' =
      Godi_plan.sort_packages to_build in

    (* Fetch packages: *)

    List.iter
      (fun pkg ->
	 let p = any_pkg pkg in
  	 pg # put_line ("### Fetching package " ^ p#name ^ " (if necessary)");
	 Godi_perform.fetch_pkg 
	   ~pref_user_error:true ~interactivity:ia ~protocol pkg;
	 poll();
	 if not !active then raise Godi_exec.Command_discontinued;
      )
      to_build';

    (* Remove packages: *)
    
    List.iter
      (fun pkg ->
	 let p = any_pkg pkg in
	 if (installed_pkg pkg)#preserve then (
	   pg # put_line ("### Package " ^ p#name ^ " cannot be removed, skipping");
	 )
	 else (
	   pg # put_line ("### Removing package " ^ p#name);
	   Godi_perform.remove_pkg 
	     ~pref_user_error:true ~interactivity:ia ~protocol pkg;
	 );
	 p#set_user_status `Keep;
	 write_wishes();
	 poll();
	 if not !active then raise Godi_exec.Command_discontinued;
      )
      to_remove';
    
    (* Build packages: *)
    
    (* Update godi.conf before build starts: *)
    let mflag = 
      Godi_conffile.modify_godi_conf ~pref_user_error:true to_build' in
    if mflag then forget_src_pkgs_on_quit := true;
    
    List.iter
      (fun pkg ->
	 let p = any_pkg pkg in
  	 pg # put_line ("### Building package " ^ p#name);
	 ( try
	     Godi_perform.build_pkg 
	       ~pref_user_error:true ~interactivity:ia ~protocol pkg;
	     p#set_user_status `Keep;
	     write_wishes();
	   with
	     | Restart_recommended _ as exn ->
		 (* Remove the package from the wish list. Note, however, that
		  * we do not know whether the installation will be successful.
		  * In this case we will have problems anyway, so it is not
		  * worth to handle this case specially.
		  *)
		 p#set_user_status `Keep;
		 write_wishes();
		 raise exn
	 );
	 poll();
	 if not !active then raise Godi_exec.Command_discontinued;
      )
      to_build';

    pg # put_line "";
    pg # put_line "### Installation completed";
    ui_continue();
  with
      Godi_exec.Command_discontinued ->
	pg # put_line "";
	pg # put_line "### Installation aborted";
	ui_continue();
    | User_error msg ->
	pg # put_line "";
	pg # put_line ("### Error: " ^ msg);
	ui_continue();
    | Restart_recommended sh_cmd ->
	pg # put_line "";
	pg # put_line "### Self-updating:";
	pg # put_line "    This package contains a replacement for the currently running program.";
	pg # put_line "    In order to perform the update properly it is now required to interrupt";
	pg # put_line "    godi_console, and to let an external process update the program, followed";
	pg # put_line "    by a restart. There is a script automating this action. It is now started.";
	ui_self_update sh_cmd;
;;


let display (pg : Miniui.page) src_list inst_list =
  
  pg # clear();

  Menu_aux.add_compact_head ~lmargin:6 pg ">>> Installation <<<";

  let corr_list = correlate (src_list @ inst_list) in
  let n_remove  = count_remove    corr_list in
  let n_upgrade = count_upgrade   corr_list in
  let n_new     = count_new_build corr_list in

  (* Check whether there are unconfigured packages: *)
  let unconf = ref [] in
  List.iter 
    (fun pkg ->
       match pkg with
	   `Source_pkg sp ->
	     ( match sp # user_status with
		   `Build
		 | `Build_dep _ ->
		     let have_config = 
		       sp # build_options <> [] || sp # build_message <> "" in
		     if have_config && not(sp # config_dialog_shown) then
		       unconf := sp :: !unconf
		 | _ -> ()
	     )
	 | _ ->
	     ()
    )
    src_list;

  pg # put_line (sprintf "GODI is going to");
  pg # put_line (sprintf " - remove %d package(s)" n_remove);
  pg # put_line (sprintf " - upgrade %d package(s) from source" n_upgrade);
  pg # put_line (sprintf " - build %d package(s) from source" n_new);

  pg # put_line "";
  pg # put_line "Press [ok] to execute this plan.";

  if !unconf <> [] then (
    pg # put_line "";
    pg # put_line "Important note: There are packages with configuration";
    pg # put_line "options or build messages you might not yet have seen:";
    List.iter
      (fun sp ->
	 pg # put_line "";
	 pg # put_line ("Package " ^ sp#name ^ ":");
	 if sp#build_options <> [] then
	   pg # put_line("  Options: " ^ String.concat ", " sp#build_options);
	 if sp#build_message <> "" then (
	   pg # put_line("  Build message:");
	   let lines = Str.split lf_re sp#build_message in
	   List.iter 
	     (fun l -> pg # put_line ("  " ^ l))
	     lines;
	 );
	 pg # put_line "";
	 pg # put_line "To set options, exit this menu, select the packages";
	 pg # put_line "in question, and invoke their [c]onfigure menu.";
      )
      !unconf;
  );

  forget_src_pkgs_on_quit := false;

  ignore(pg # add_foot_button 
	   ~text:"ok"
	   ~underline:0
	   ~callback:(fun () -> 
			( try
			    execute pg src_list inst_list;
			  with
			      User_error msg ->
				Menu_aux.error_page pg msg
			);
			(* Force rescan of the installed packages: *)
			Godi_query.forget_installed_packages();
			())
	   (`Keypress 'o'));

  ignore(pg # add_foot_button ~text:"help" ~underline:0
           ~callback:(Menu_help.display pg#new_page help_str)
           (`Keypress 'h'));

  ignore(pg # add_foot_button 
	   ~text:"exit menu"
	   ~underline:1
	   ~callback:(fun () -> pg # quit())
	   (`Keypress 'x'));

  pg # loop();

  (* Additionally, we also force the rescan of the source packages, if
   * there have been changes of the options.
   *)
  if !forget_src_pkgs_on_quit then
    Godi_query.forget_source_packages();

;;
