(* $Id: godi_perform.ml 426 2004-05-07 22:00:28Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Printf

let fetch_pkg ?(pref_user_error=false) ?interactivity ?protocol pkg =
  if pkgtype pkg <> `Source then
    failwith "Godi_perform.fetch_pkg: not a source package";
  let dir = (source_pkg pkg) # source_directory in
  exec_to_user_error
    ~pref_user_error
    (Godi_exec.call_cmd
       ~cd:dir
       ?stdout:protocol
       ?stderr:protocol
       ?interactivity)
    ("godi_make", [| "fetch" |] )
;;


let remove_pkg ?(pref_user_error=false) ?interactivity ?protocol pkg =
  if pkgtype pkg <> `Installed then
    failwith "Godi_perform.remove_pkg: not an installed package";
  exec_to_user_error
    ~pref_user_error
    (Godi_exec.call_cmd
       ?stdout:protocol
       ?stderr:protocol
       ?interactivity)
    ("godi_delete", [| pkg_name pkg |] )
;;


let add_pkg ?(pref_user_error=false) ?interactivity ?protocol pkg =
  if pkgtype pkg <> `Binary then
    failwith "Godi_perform.add_pkg: not a binary package";
  exec_to_user_error
    ~pref_user_error
    (Godi_exec.call_cmd
       ?stdout:protocol
       ?stderr:protocol
       ?interactivity)
    ("godi_add", [| (binary_pkg pkg)#binary_archive |] )
;;


let test_conf_pkg ?(pref_user_error=false) ?interactivity ?protocol
                  ?(godi_conf = Godi_query.godi_conf()) pkg =
  if pkgtype pkg <> `Source then
    failwith "Godi_perform.test_conf_pkg: not a source package";
  let dir = (source_pkg pkg) # source_directory in
  exec_to_user_error
    ~pref_user_error
    (fun () ->
       Godi_exec.call_cmd
       ~cd:dir
       ?stdout:protocol
       ?stderr:protocol
       ?interactivity
       ("godi_make", [| "clean" |] );
       (* Classic source package *)
       Godi_exec.call_cmd
	 ~cd:dir
	 ~env:[ "GODI_CONF", godi_conf ]
	 ?stdout:protocol
	 ?stderr:protocol
	 ?interactivity
	 ("godi_make", [| "configure" |] );
       Godi_exec.call_cmd
	 ~cd:dir
	 ?stdout:protocol
	 ?stderr:protocol
	 ?interactivity
	 ("godi_make", [| "clean" |] )
    )
    ()
;;


let build_pkg ?(pref_user_error=false) ?interactivity ?protocol 
              ?(catch_self = Some "godi-tools") pkg =
  if pkgtype pkg <> `Source then
    failwith "Godi_perform.build_pkg: not a source package";
  let dir = (source_pkg pkg) # source_directory in
  exec_to_user_error
    ~pref_user_error
    (fun () ->
       (* TODO:
	* - catch_self is only applied when the package is already installed.
	*   Currently godi_delete outputs an error, but the error is ignored.
	*)
       Godi_exec.call_cmd
         ~cd:dir
         ?stdout:protocol
         ?stderr:protocol
         ?interactivity
         ("godi_make", [| "clean" |] );
       if (source_pkg pkg)#self_update_pkg <> "" then (
	 (* Package to self-update GODI: *)
	 Godi_exec.call_cmd
	   ~cd:dir
	   ?stdout:protocol
	   ?stderr:protocol
	   ?interactivity
	   ("godi_make", [| "self_package" |] );
	 (* Sleep here for a second, to work around a bug in godi_tar.
	  * This little pause can go away when it is unlikely that old
	  * versions of godi_tar are around.
	  *)
	 Unix.sleep 1;
	 (* If this package is also meant by catch_self, we do not replace
	  * the package immediately. Instead, we exit the program now and
	  * execute shell commands replacing the package followed by a
	  * restart of the program.
	  *)
	 if (catch_self = Some (source_pkg pkg)#name) then (
	   Godi_exec.call_cmd
	      ~cd:dir
	      ?stdout:protocol
   	      ?stderr:protocol
	      ?interactivity
	      ("godi_make", [| "clean" |] );
	   let sh_cmd = 
	     sprintf "godi_delete -f %s\ngodi_add %s"
	             (Filename.quote (source_pkg pkg)#name)
	             (Filename.quote (source_pkg pkg)#self_update_pkg) in
	   raise (Restart_recommended sh_cmd)
	 );
	 (* In the normal case, run these commands directly: *)
	 (* Force deletion, ignore errors: *)
	 ( try
	     Godi_exec.call_cmd
	       ?stdout:protocol
	       ?stderr:protocol
	       ?interactivity
	       ("godi_delete", [| "-f"; (source_pkg pkg)#name |] )
	   with
	       Godi_exec.Command_failure(_,_,_) ->
		 ()
	 );
	 Godi_exec.call_cmd
	   ?stdout:protocol
	   ?stderr:protocol
	   ?interactivity
	   ("godi_add", [| (source_pkg pkg)#self_update_pkg |] )
       ) else (
	 (* Classic source package *)
	 if (catch_self = Some (source_pkg pkg)#name) then (
	   (* We cannot handle this case. Output a cryptic warning: *)
	   (match protocol with
		Some p ->
		  Godi_exec.collect p "Warning: This package may fail to replace itself\n";
	      | None -> ()
	   )
	 );
	 Godi_exec.call_cmd
		 ~cd:dir
		 ?stdout:protocol
		 ?stderr:protocol
		 ?interactivity
		 ("godi_make", [| "package" |] );
       );
       Godi_exec.call_cmd
	 ~cd:dir
	 ?stdout:protocol
	 ?stderr:protocol
	 ?interactivity
	 ("godi_make", [| "clean" |] );
    )
    ()
;;

