(* $Id: godi_types.mli 300 2004-03-17 22:14:28Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* Common types for GODI tools *)

open Godi_versions
open Godi_exec

exception User_error of string
  (* An non-fatal error message that should be presented to the user *)


val user_error : string -> 'a
  (* Raises User_error *)

val exec_to_user_error : ?pref_user_error:bool -> ('a -> 'b) -> 'a -> 'b
  (* Calls the function with the argument and returns the result.
   * If pref_user_error=true, the Godi_exec exceptions are caught
   * and transformed to the corresponding User_error exceptions.
   * If pref_user_error=false (the default), the exceptions are not
   * caught.
   *)


exception Restart_recommended of string
  (* Indicates that the program should now be restarted. Used by some
   * functions as special result.
   * The string argument is a special script that should be executed
   * from outside of godi_console. This script does not include the
   * command to restart godi_console (in contrast to the Restart exception).
   *)

exception Restart of string
  (* This exception falls through to the main program. The string argument
   * is a script that is executed by /bin/sh. This script can do any
   * actions, but finally an exec of the godi_console executable should
   * be done.
   *)


type user_status =
    [ `Keep        (* Keep installed or deinstalled, whatever applies *)
    | `Build       (* Build because of user's wish *)
    | `Remove      (* Remove because of user's wish *)
    | `Build_dep of int  
                   (* Build because of dependencies. The int is the iteration
		    * number of the user dialog
		    *)
    | `Remove_dep of int
	           (* Remove because of dependencies. The int is the iteration
		    * number of the user dialog
		    *)
    (* | `Add      (* Add binary package *) *)
    ]
;;

(* What are iteration numbers?
 *
 * godi_console often displays the package list again after the
 * dependency analysis because packages have been selected or deselected.
 * The iteration number is increased when the list is shown again,
 * such that packages whose dependency status change can be 
 * distinguished from packages whose status has changed in one of
 * previous cycles.
 *)


type var_id 
  (* The variant ID, used to distinguish between objects describing
   * the same package under different conditions.
   *)


val new_var_id : unit -> var_id
  (* allocates a new variant ID *)


class type any_package =
object ('self)
  method name : string
  method version : string
    (* Note: version always without pkgrevision, i.e. no "godi" suffix *)
  method pkgrevision : int
    (* The pkgrevision is 0 if the full version string has no "godi"
     * suffix, and it is > 0 if the full version string has a "godi"
     * suffix.
     *)

  method comment : string
  method description : string

  method dependencies : dependency list
  method build_dependencies : dependency list

  (* not yet: conflicts *)

  method user_status : user_status
  method set_user_status : user_status -> unit
    (* The user_status is not a persistent property. After the packages
     * have been scanned, the user_status is always `Keep. Use the 
     * functions in Godi_wishes to apply the wish list.
     *)

  method var_id : var_id
    (* The [var_id] of this object *)
end


class type mixin_variant =
object('self)
  method variant : var_id -> 'self
    (* Returns a copy of this object for the demanded [var_id]. The
     * method fails if the [var_id] is unknown, or not applicable to
     * this object.
     *)
end


class type source_package =
object
  inherit any_package
  inherit mixin_variant

  method source_directory : string
    (* Absolute path of the directory with the driving "Makefile", i.e.
     * <prefix>/build/<cat>/<package>
     *)

  method build_message : string

  method build_options : string list
    (* Returns the names of the configure options for godi.conf (i.e.
     * the contents of CONFOPTS)
     *)

  method values_of_build_options : (string * string) list
    (* Returns also the values of the options. The options are in the
     * same order as in the list returned by [build_options].
     *)

  method values_of_user_build_options : (string * string) list
    (* Returns the values as set by [set_user_build_options], or
     * the same as [values_of_build_options] if this method has not
     * yet been called.
     *)

  method set_user_build_options : (string * string) list -> unit
    (* This method sets the user's wishes of the option values.
     * It is required that List.map fst arg is equal to
     * the result of the method [build_options].
     *)

  method config_dialog_shown : bool
  method set_config_dialog_shown : bool -> unit
    (* The flag [config_dialog_shown] is set when the configure dialog
     * was shown to the user
     *)

  method self_update_pkg : string
    (* If this string is non-empty, the package must be handled in
     * a special way for installation. Instead of "bmake package",
     * the bmake target "self_package" must be called. After that,
     * there is a binary package file where [self_update_pkg] 
     * specifies, i.e. the method returns the absolute path of the
     * archive file. This binary package should be installed.
     *
     * The value of this method corresponds to the make variable
     * SELF_UPDATE_PKG. Do not use this feature for user packages,
     * it is reserved for GODI to update itself.
     *)

  method rescan_deps : ?protocol:collector -> string -> var_id -> unit
    (* Rescans the values for DEPENDS and BUILD_DEPENDS for a different
     * godi.conf file. The name of this file must be passed as string.
     * The [var_id] is the variant ID that can be the argument of
     * [variant] to get the object that corresponds to the changed
     * settings, i.e. the methods [dependencies] and [build_dependencies]
     * of the variant return the new values.
     *)
end


class type binary_package =
object
  inherit any_package
  inherit mixin_variant

  method binary_archive : string
    (* The absolute path to the binary archive (the .tgz file) *)

end


class type installed_package =
object
  inherit any_package
  inherit mixin_variant

  method preserve : bool
    (* A "preserved" package cannot be deleted *)

end


type package =
    [ `Source_pkg of source_package
    | `Binary_pkg of binary_package
    | `Installed_pkg of installed_package
    ]

type pkgtype =
    [ `Source
    | `Binary
    | `Installed
    ]


val source_pkg : package -> source_package
val binary_pkg : package -> binary_package
val installed_pkg : package -> installed_package
  (* These functions fail if the package is not of the assumed type *)

val any_pkg : package -> any_package
val pkgtype : package -> pkgtype

(* Abbreviations: e.g. pkg_name p <=> (any_pkg p)#name *)
val pkg_name               : package -> string
val pkg_version            : package -> string
val pkg_pkgrevision        : package -> int
val pkg_comment            : package -> string
val pkg_description        : package -> string
val pkg_dependencies       : package -> dependency list
val pkg_build_dependencies : package -> dependency list
val pkg_user_status        : package -> user_status

val to_be_built : package -> bool
  (* If user_status is `Build or `Build_dep *)

val to_be_removed : package -> bool
  (* If user_status is `Remove or `Remove_dep *)

val newer_package : package -> package -> bool
  (* Whether the first package is newer than the second package.
   * The package revision numbers are taken into account.
   *)

val correlate : package list -> 
                  (string * 
		   package list * 
		   package list * 
		   package list) list
  (* For every package name, the returned list contains the
   * tuple (name, srclist, binlist, instlist) where
   * srclist is the list of all source packages with this name,
   * binlist is the list of all binary packages with this name,
   * instlist is the list of all installed packages with this name.
   *
   * Furthermore, the returned list is sorted by names (ascending),
   * but the sublists may be in any order.
   *)


module Pkg : sig
  (* To create sets of packages: Set.Make(Godi_types.Pkg) *)
  type t = package
  val compare : t -> t -> int
    (* Compares the object identifiers of the packages *)
end


module PkgSet : Set.S with type elt = Pkg.t
  (* Sets of packages, based on object identification *)


module PkgMap : sig
  (* An improved and specialized version of Map: This version supports
   * querying the elements of the mapping by the properties name, version,
   * pkgrevision, and pkgtype. Furthermore, there are also some set methods,
   * so it may be helpful to consider "unit PkgMap.t" as a set implementation.
   *)

  type (+'a) t
  type key = package

  val empty : 'a t                        (* as in Map *)
  val is_empty : 'a t -> bool             (* faster than x = empty *)
  val add: key -> 'a -> 'a t -> 'a t      (* as in Map *)
  val find: key -> 'a t -> 'a             (* as in Map *)
  val find_by : ?name:string -> 
                ?version:string ->
		?pkgrevision:int ->
		?pkgtype:pkgtype ->
                'a t -> 
		  (key * 'a) list
    (* Returns all elements that match the criterions ~name, ~version,
     * ~pkgrevision, ~pkgtype, if present. Without search criterion,
     * all packages are returned.
     *)
  val remove: key -> 'a t -> 'a t
  val mem: key -> 'a t -> bool            (* as in Map *)
  val mem_by : ?name:string -> 
               ?version:string ->
	       ?pkgrevision:int ->
	       ?pkgtype:pkgtype ->
               'a t -> 
		  bool
    (* Checks whether there are elements matching the criterion. *)

  val iter: (key -> 'a -> unit) -> 'a t -> unit          (* as in Map *)
  val map: ('a -> 'b) -> 'a t -> 'b t                    (* as in Map *)
  val mapi: (key -> 'a -> 'b) -> 'a t -> 'b t            (* as in Map *)
  val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b  (* as in Map *)

  (* Interoperability with lists: *)
  val make: (key * 'a) list -> 'a t
  val list: 'a t -> (key * 'a) list
  val keys: 'a t -> key list
  val values: 'a t -> 'a list

  (* Sets: *)
  val union : 'a t -> 'a t -> 'a t
  val diff : 'a t -> 'a t -> 'a t
  val inter : 'a t -> 'a t -> 'a t

  val for_all : (key -> 'a -> bool) -> 'a t -> bool
  val exists :  (key -> 'a -> bool) -> 'a t -> bool
  val filter_list :  (key -> 'a -> bool) -> 'a t -> (key * 'a) list
  val filter :  (key -> 'a -> bool) -> 'a t -> 'a t
  val partition_list : (key -> 'a -> bool) -> 'a t -> 
                         (key * 'a) list * (key * 'a) list
  val partition : (key -> 'a -> bool) -> 'a t -> 'a t * 'a t

end
